# -*- coding: utf-8 -*-
from aigpy import cmdHelper
from aigpy import systemHelper

def printErr(length, elsestr):
    cmdHelper.myprint("[ERR]".ljust(length), cmdHelper.TextColor.Red, None)
    print(elsestr)


def printWarning(length, elsestr):
    cmdHelper.myprint("[WARNING]".ljust(length), cmdHelper.TextColor.Red, None)
    print(elsestr)

def printInfo(length, elsestr):
    cmdHelper.myprint("[INFO]".ljust(length), cmdHelper.TextColor.Yellow, None)
    print(elsestr)

def printSUCCESS(length, elsestr):
    cmdHelper.myprint("[SUCCESS]".ljust(length), cmdHelper.TextColor.Green, None)
    print(elsestr)

def printChoice(string, isInt=False, default=None):
    tmpstr = ""
    cmdHelper.myprint(string, cmdHelper.TextColor.Yellow, None)
    if not isInt:
        return cmdHelper.myinput(tmpstr)
    else:
        return cmdHelper.myinputInt(tmpstr, default)
def printChoice2(string, default=None):
    ret = printChoice(string, False, default)
    try:
        iret = int(ret)
        return ret, iret
    except:
        return ret, default
def printMenu():
    print("=====================Choice=========================")
    cmdHelper.myprint(" Enter '0' : ", cmdHelper.TextColor.Green, None)
    print("Exit.")
    cmdHelper.myprint(" Enter '1' : ", cmdHelper.TextColor.Green, None)
    print("LogIn And Get SessionID.")
    cmdHelper.myprint(" Enter '2' : ", cmdHelper.TextColor.Green, None)
    print("Change Settings.")
    cmdHelper.myprint(" Enter '3' : ", cmdHelper.TextColor.Green, None)
    print("Download an Album.")
    cmdHelper.myprint(" Enter '4' : ", cmdHelper.TextColor.Green, None)
    print("Download a Track.")
    cmdHelper.myprint(" Enter '5' : ", cmdHelper.TextColor.Green, None)
    print("Download a PlayList.")
    cmdHelper.myprint(" Enter '6' : ", cmdHelper.TextColor.Green, None)
    print("Download a Video.")
    cmdHelper.myprint(" Enter '7' : ", cmdHelper.TextColor.Green, None)
    print("Download a Favorite.")
    cmdHelper.myprint(" Enter '8' : ", cmdHelper.TextColor.Green, None)
    print("Download All Albums by an Artist.")
    cmdHelper.myprint(" Enter '9' : ", cmdHelper.TextColor.Green, None)
    print("Show Current Config.")
    cmdHelper.myprint(" Enter URL : ", cmdHelper.TextColor.Green, None)
    print("Download By URL.")
    cmdHelper.myprint(" Enter Path: ", cmdHelper.TextColor.Green, None)
    print("Download By File.")
    print("====================================================")
    

LOG = '''
 /$$$$$$$$ /$$       /$$           /$$               /$$ /$$
|__  $$__/|__/      | $$          | $$              | $$| $$
   | $$    /$$  /$$$$$$$  /$$$$$$ | $$          /$$$$$$$| $$
   | $$   | $$ /$$__  $$ |____  $$| $$ /$$$$$$ /$$__  $$| $$
   | $$   | $$| $$  | $$  /$$$$$$$| $$|______/| $$  | $$| $$
   | $$   | $$| $$  | $$ /$$__  $$| $$        | $$  | $$| $$
   | $$   | $$|  $$$$$$$|  $$$$$$$| $$        |  $$$$$$$| $$
   |__/   |__/ \_______/ \_______/|__/         \_______/|__/
   
       https://github.com/yaronzz/Tidal-Media-Downloader 
'''



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
'''
@File    :   download.py
@Time    :   2019/02/27
@Author  :   Yaron Huang
@Version :   1.0
@Contact :   yaronhuang@qq.com
@Desc    :   Download Function
'''
import sys
import os
import codecs

from datetime import datetime
from aigpy import pathHelper
# from tidal_dl import netHelper
from aigpy import netHelper
from aigpy import fileHelper
from aigpy import cmdHelper
from aigpy import systemHelper

# from tidal_dl.ffmpegHelper import FFmpegTool
from aigpy.ffmpegHelper import FFmpegTool
from aigpy.cmdHelper import myinput, myinputInt
from aigpy.threadHelper import ThreadTool
from aigpy.progressHelper import ProgressTool

from tidal_dl.check import CheckTool
from tidal_dl.tidal import TidalTool
from tidal_dl.tidal import TidalConfig
from tidal_dl.tidal import TidalAccount
from tidal_dl.decryption import decrypt_security_token
from tidal_dl.decryption import decrypt_file
from tidal_dl.printhelper import printChoice, printErr, printSUCCESS, printWarning, printInfo


class Download(object):
    def __init__(self, threadNum=3):
        self.config = TidalConfig()
        self.tool = TidalTool()
        self.thread = ThreadTool(int(threadNum))
        self.ffmpeg = FFmpegTool(mergerTimeout=45)
        self.progress = ProgressTool(100)
        self.check = CheckTool()

        self.showpro = False
        if self.config.showprogress == 'True':
            self.showpro = True

        pathHelper.mkdirs(self.config.outputdir + "/Album/")
        pathHelper.mkdirs(self.config.outputdir + "/Playlist/")
        pathHelper.mkdirs(self.config.outputdir + "/Video/")
        pathHelper.mkdirs(self.config.outputdir + "/Favorite/")

    def __isNeedDownload(self, path, url):
        curSize = fileHelper.getFileSize(path)
        if curSize <= 0:
            return True
        netSize = netHelper.getFileSize(url)
        if curSize >= netSize:
            return False
        return True

    # dowmload track thread
    def __thradfunc_dl(self, paraList):
        count = 1
        printRet = True
        pstr = paraList['title'] + "(Download Err!)"
        redownload = True
        needDl = True
        bIsSuccess = False
        albumInfo = None
        index = None
        coverpath = None

        if 'redownload' in paraList:
            redownload = paraList['redownload']
        if 'retry' in paraList:
            count = count + paraList['retry']
        if 'show' in paraList:
            printRet = paraList['show']
        if 'album' in paraList:
            albumInfo = paraList['album']
        if 'index' in paraList:
            index = paraList['index']
        if 'coverpath' in paraList:
            coverpath = paraList['coverpath']

        if redownload is False:
            needDl = self.__isNeedDownload(paraList['path'], paraList['url'])

        # DEBUG
        # self.tool.setTrackMetadata(paraList['trackinfo'], paraList['path'], albumInfo, index, coverpath)
        showprogress = False
        if int(self.config.threadnum) <= 1 and self.showpro:
            showprogress = True

        Contributors = self.tool.getTrackContributors(paraList['trackinfo']['id'])
        if needDl:
            try:
                while count > 0:
                    count = count - 1
                    check = netHelper.downloadFile(paraList['url'], paraList['path']+'.part', showprogress=showprogress, stimeout=20)
                    if check is True:
                        if paraList['key'] == '':
                            # unencrypted -> just move into place
                            os.replace(paraList['path']+'.part', paraList['path'])
                            break
                        else:
                            # encrypted -> decrypt and remove encrypted file
                            key, nonce = decrypt_security_token(paraList['key'])
                            decrypt_file(paraList['path']+'.part', paraList['path'], key, nonce)
                            os.remove(paraList['path']+'.part')
                        break
                if check:
                    bIsSuccess = True
                    if self.tool.isNeedCovertToM4a(paraList['path']):
                        if paraList['codec'] == 'ac4':
                            printInfo(14, 'Skip convert to m4a(AC4-Codec).')
                        elif paraList['codec'] == 'mha1':
                            printInfo(14, 'Skip convert to m4a(MHA1-Codec).')
                        else:
                            paraList['path'] = self.tool.covertMp4toM4a(paraList['path'])
                    self.tool.setTrackMetadata(paraList['trackinfo'], paraList['path'],
                                               albumInfo, index, coverpath, Contributors)
                    pstr = paraList['title']
            except Exception as e:
                printErr(14, str(e) + " while downloading " + paraList['url'])
        else:
            pstr = paraList['title']
            bIsSuccess = True

        if printRet:
            if(bIsSuccess):
                printSUCCESS(14, pstr)
            else:
                printErr(14, pstr)
        return

    # creat album output dir
    def __creatAlbumDir(self, albumInfo, quality='LOW'):
        # creat outputdir
        title = pathHelper.replaceLimitChar(albumInfo['title'], '-')
        author = pathHelper.replaceLimitChar(albumInfo['artist']['name'], '-')

        # add year
        if self.config.addyear != 'No':
            if self.config.addyear == 'Before':
                title = '[' + str(datetime.strptime(albumInfo['releaseDate'], '%Y-%m-%d').year) + '] '+title
            elif self.config.addyear == 'After':
                title = title+' [' + str(datetime.strptime(albumInfo['releaseDate'], '%Y-%m-%d').year) + ']'
            else:
                title = title
        
        # add quality[M] labels
        if 'audioQuality' in albumInfo and albumInfo['audioQuality'] == 'HI_RES' and quality == 'HI_RES':
            title = '[M] '+title

        targetDir = self.config.outputdir + "/Album/" + author + '/' + title
        targetDir = os.path.abspath(targetDir)
        pathHelper.mkdirs(targetDir)

        # creat volumes dir
        count = 1
        numOfVolumes = int(albumInfo['numberOfVolumes'])
        if numOfVolumes > 1:
            while count < numOfVolumes + 1:
                volumeDir = targetDir + "/Volume" + str(count)
                pathHelper.mkdirs(volumeDir)
                count = count + 1
        return targetDir

    def _getSongExtension(self, downloadUrl):
        if downloadUrl.find('.flac?') != -1:
            return '.flac'
        if downloadUrl.find('.m4a?') != -1:
            return '.m4a'
        if downloadUrl.find('.mp4?') != -1:
            return '.mp4'
        return '.m4a'

    def _IsExplicitString(self, IsExplicit):
        String = None
        if IsExplicit:
            String = 'Explicit'
        return String

    def __getAlbumSongSavePath(self, targetDir, albumInfo, item, extension):
        if extension is None:
            extension = ".m4a"

        seq = self.tool.getIndexStr(item['trackNumber'], albumInfo['numberOfTracks'])
        name = seq + pathHelper.replaceLimitChar(item['title'], '-')
        fileExplicit = self._IsExplicitString(item['explicit'])
        if self.config.addhyphen == 'True':
            name = seq + '- ' + pathHelper.replaceLimitChar(item['title'], '-')
        if self.config.addexplicit == "True" and fileExplicit is not None:
            name = name + " - " + fileExplicit

        seq = item['volumeNumber']
        path = targetDir + "/"
        if int(albumInfo['numberOfVolumes']) > 1:
            path += 'Volume' + str(seq) + "/"

        maxlen = 255
        if systemHelper.isLinux():
            maxlen = 4090
        # truncate filename when it's longer than system's
        # filename limit which is 255
        len_sum = len(path) + len(name) + len(extension)
        if len_sum > maxlen:
            diff = maxlen - len_sum
            name = name[: len(name) + diff]

        filePath = path + name + extension
        checklen = len(filePath)
        return filePath

    def __getExistFiles(self, paths):
        ret = []
        for item in paths:
            if os.path.isfile(item):
                ret.append(item)
        return ret

    def __getVideoResolutionIndex(self, reslist):
        array = []
        # if reslist != None:
        #     for item in reslist:
        #         subs = item.split('x')
        #         subs = subs[1].split(',')
        #         array.append(int(subs[0]))
        for item in reslist:
            subs = item.split('x')
            subs = subs[1].split(',')
            array.append(int(subs[0]))
        cmp = int(self.config.resolution)
        ret = 0
        for item in array:
            if cmp >= item:
                return ret
            ret += 1
        return len(array) - 1

    def downloadAlbum(self, album_id=None, redl_flag=None):
        while_count = 9999
        while while_count > 0:
            while_count -= 1

            if album_id is not None:
                while_count = 0
                sID = album_id
            else:
                print("----------------ALBUM------------------")
                sID = printChoice("Enter AlbumID(Enter '0' go back):", True, 0)
                if sID == 0:
                    return

            aAlbumInfo = self.tool.getAlbum(sID)
            if self.tool.errmsg != "":
                printErr(0, "Get AlbumInfo Err! " + self.tool.errmsg)
                continue

            print("[Title]       %s" % (aAlbumInfo['title']))
            print("[SongNum]     %s\n" % (aAlbumInfo['numberOfTracks']))

            # Get Tracks
            aAlbumTracks = self.tool.getAlbumTracks(sID)
            if self.tool.errmsg != "":
                printErr(0, "Get AlbumTracks Err!" + self.tool.errmsg)
                continue
            aAlbumVideos = self.tool.getAlbumVideos(sID)

            # Creat OutputDir
            targetDir = self.__creatAlbumDir(aAlbumInfo, self.config.quality)
            # write msg
            string = self.tool.convertAlbumInfoToString(aAlbumInfo, aAlbumTracks)
            with codecs.open(targetDir + "/AlbumInfo.txt", 'w', 'utf-8') as fd:
                fd.write(string)
            # download cover
            coverPath = targetDir + '/' + pathHelper.replaceLimitChar(aAlbumInfo['title'], '-') + '.jpg'
            if aAlbumInfo['cover'] is not None:
                coverUrl = self.tool.getAlbumArtworkUrl(aAlbumInfo['cover'])
                netHelper.downloadFile(coverUrl, coverPath)
            # check exist files
            redownload = True
            if redl_flag is None:
                existFiles = pathHelper.getDirFiles(targetDir)
                for item in existFiles:
                    if '.txt' in item:
                        continue
                    if '.jpg' in item:
                        continue
                    check = printChoice("Some tracks already exist. Redownload?(y/n):")
                    if not cmdHelper.isInputYes(check):
                        redownload = False
                    break
            else:
                redownload = redl_flag

            # download album tracks
            for item in aAlbumTracks['items']:
                streamInfo = self.tool.getStreamUrl(str(item['id']), self.config.quality)
                if self.tool.errmsg != "" or not streamInfo:
                    printErr(14, item['title'] + "(Get Stream Url Err!" + self.tool.errmsg + ")")
                    continue

                fileType = self._getSongExtension(streamInfo['url'])
                filePath = self.__getAlbumSongSavePath(targetDir, aAlbumInfo, item, fileType)
                paraList = {'album': aAlbumInfo,
                            'redownload': redownload,
                            'title': item['title'],
                            'trackinfo': item,
                            'url': streamInfo['url'],
                            'path': filePath,
                            'retry': 3,
                            'key': streamInfo['encryptionKey'],
                            'coverpath': coverPath,
                            'codec': streamInfo['codec']}
                self.thread.start(self.__thradfunc_dl, paraList)
            # wait all download thread
            self.thread.waitAll()
            self.tool.removeTmpFile(targetDir)
            
            # remove cover
            if self.config.savephoto != 'True':
                pathHelper.remove(coverPath)

            # download video

            for item in aAlbumVideos:
                item = item['item']
                filePath = targetDir + '/' + pathHelper.replaceLimitChar(item['title'], '-') + ".mp4"
                filePath = os.path.abspath(filePath)
                if os.access(filePath, 0):
                    os.remove(filePath)

                try:
                    resolutionList, urlList = self.tool.getVideoResolutionList(item['id'])
                    selectIndex = self.__getVideoResolutionIndex(resolutionList)
                    if self.ffmpeg.mergerByM3u8_Multithreading2(urlList[int(selectIndex)], filePath, showprogress=self.showpro):
                        printSUCCESS(14, item['title'])
                    else:
                        printErr(14, item['title'])
                except:
                    printErr(14, item['title'])
            # return

        return

    def downloadArtistAlbum(self, includeSingles=True, artistID=None):
        while True:
            print("-------------ARTIST ALBUM--------------")
            if artistID is not None:
                sID = artistID
            else:
                sID = printChoice("Enter Artist ID(Enter '0' go back):", True, 0)
                if sID == 0:
                    return

            array = self.tool.getArtistAlbum(sID, includeSingles)
            if self.tool.errmsg != "":
                printErr(0, "Get AlbumList Err! " + self.tool.errmsg)
                continue

            redownload = True
            if artistID is None:
                check = printChoice("Skip downloaded files?(y/n):")
                if cmdHelper.isInputYes(check):
                    redownload = False

            for index, item in enumerate(array):
                print("----Album[{0}/{1}]----".format(index+1, len(array)))
                self.downloadAlbum(item['id'], redownload)

            if artistID is not None:
                # Break out of the function if we are only downloading one artist's albums
                return



    def downloadTrack(self, track_id=None):
        while_count = 9999
        while while_count > 0:
            while_count -= 1

            if track_id is not None:
                while_count = 0
                sID = track_id
            else:
                print("----------------TRACK------------------")
                sID = printChoice("Enter TrackID(Enter '0' go back):", True, 0)
                if sID == 0:
                    return
            aTrackInfo = self.tool.getTrack(sID)
            if self.tool.errmsg != "":
                printErr(0, "Get TrackInfo Err! " + self.tool.errmsg)
                return
            aAlbumInfo = self.tool.getAlbum(aTrackInfo['album']['id'])
            if self.tool.errmsg != "":
                printErr(0, "Get TrackInfo Err! " + self.tool.errmsg)
                return

            # t = self.tool.getTrackContributors(sID)

            print("[AlbumTitle ]       %s" % (aAlbumInfo['title']))
            print("[TrackTitle ]       %s" % (aTrackInfo['title']))
            print("[Duration   ]       %s" % (aTrackInfo['duration']))
            print("[TrackNumber]       %s" % (aTrackInfo['trackNumber']))
            print("[Explicit   ]       %s" % (aAlbumInfo['explicit']))
            # print("[Version    ]       %s\n" % (aTrackInfo['version']))

            # Creat OutputDir
            targetDir = self.__creatAlbumDir(aAlbumInfo, self.config.quality)
            # download cover
            coverPath = targetDir + '/' + pathHelper.replaceLimitChar(aAlbumInfo['title'], '-') + '.jpg'
            if aAlbumInfo['cover'] is not None:
                coverUrl = self.tool.getAlbumArtworkUrl(aAlbumInfo['cover'])
                netHelper.downloadFile(coverUrl, coverPath)

            # download
            streamInfo = self.tool.getStreamUrl(sID, self.config.quality)
            if self.tool.errmsg != "" or not streamInfo:
                printErr(14, aTrackInfo['title'] + "(Get Stream Url Err!" + self.tool.errmsg + ")")
                continue
            
            print("[Codec      ]       %s" % (streamInfo['codec']))
            fileType = self._getSongExtension(streamInfo['url'])
            filePath = self.__getAlbumSongSavePath(targetDir, aAlbumInfo, aTrackInfo, fileType)
            paraList = {'album': aAlbumInfo,
                        'title': aTrackInfo['title'],
                        'trackinfo': aTrackInfo,
                        'url': streamInfo['url'],
                        'path': filePath,
                        'retry': 3,
                        'key': streamInfo['encryptionKey'],
                        'coverpath': coverPath,
                        'codec': streamInfo['codec']}

            self.thread.start(self.__thradfunc_dl, paraList)
            # wait all download thread
            self.thread.waitAll()
            self.tool.removeTmpFile(targetDir)

            # remove cover
            if self.config.savephoto != 'True':
                pathHelper.remove(coverPath)
        return

    def downloadVideo(self, video_id=None):
        flag = True
        while flag:
            targetDir = self.config.outputdir + "/Video/"
            if video_id is None:
                print("----------------VIDEO------------------")
                sID = printChoice("Enter VideoID(Enter '0' go back):", True, 0)
                if sID == 0:
                    return
            else:
                flag = False
                sID = video_id

            aVideoInfo = self.tool.getVideo(sID)
            if self.tool.errmsg != "":
                printErr(0, "Get VideoInfo Err! " + self.tool.errmsg)
                continue

            print("[Title      ]       %s" % (aVideoInfo['title']))
            print("[Duration   ]       %s" % (aVideoInfo['duration']))
            print("[TrackNumber]       %s" % (aVideoInfo['trackNumber']))
            print("[Type       ]       %s\n" % (aVideoInfo['type']))

            # get resolution
            index = 0
            resolutionList, urlList = self.tool.getVideoResolutionList(sID)
            if self.tool.errmsg != "":
                printErr(14, self.tool.errmsg)
                continue

            index = self.__getVideoResolutionIndex(resolutionList)
            path = targetDir + "/" + pathHelper.replaceLimitChar(aVideoInfo['title'], '-') + ".mp4"
            path = os.path.abspath(path)
            if os.access(path, 0):
                os.remove(path)

            if self.ffmpeg.mergerByM3u8_Multithreading2(urlList[int(index)], path, True):
                printSUCCESS(14, aVideoInfo['title'])
            else:
                printErr(14, aVideoInfo['title'])
        return

    def downloadPlaylist(self, playlist_id=None):
        while True:
            targetDir = self.config.outputdir + "/Playlist/"
            if playlist_id is None:
                print("--------------PLAYLIST-----------------")
                sID = printChoice("Enter PlayListID(Enter '0' go back):")
                if sID == '0':
                    return
            else:
                sID = playlist_id

            aPlaylistInfo, aItemInfo = self.tool.getPlaylist(sID)
            if self.tool.errmsg != "":
                printErr(0, "Get PlaylistInfo Err! " + self.tool.errmsg)
                return

            print("[Title]                %s" % (aPlaylistInfo['title']))
            print("[Type]                 %s" % (aPlaylistInfo['type']))
            print("[NumberOfTracks]       %s" % (aPlaylistInfo['numberOfTracks']))
            print("[NumberOfVideos]       %s" % (aPlaylistInfo['numberOfVideos']))
            print("[Duration]             %s\n" % (aPlaylistInfo['duration']))

            # Creat OutputDir
            targetDir = targetDir + pathHelper.replaceLimitChar(aPlaylistInfo['title'], '-')
            targetDir = os.path.abspath(targetDir).strip()
            pathHelper.mkdirs(targetDir)
            # write msg
            string = self.tool.convertPlaylistInfoToString(aPlaylistInfo, aItemInfo)
            with codecs.open(targetDir + "/PlaylistInfo.txt", 'w', 'utf-8') as fd:
                fd.write(string)
            # download cover
            coverPath = targetDir + '/' + pathHelper.replaceLimitChar(aPlaylistInfo['title'], '-') + '.jpg'
            coverUrl = self.tool.getPlaylistArtworkUrl(aPlaylistInfo['uuid'])
            check = netHelper.downloadFile(coverUrl, coverPath)

            # download track
            bBreakFlag = False
            bFirstTime = True
            errIndex = []
            index = 0

            while bBreakFlag is False:
                self.check.clear()
                index = 0
                tmpcoverpath = []
                for item in aItemInfo:
                    type = item['type']
                    item = item['item']
                    if type != 'track':
                        continue

                    index = index + 1
                    if bFirstTime is False:
                        if self.check.isInErr(index - 1, errIndex) == False:
                            continue

                    streamInfo = self.tool.getStreamUrl(str(item['id']), self.config.quality)
                    # streamInfo = self.tool.getStreamUrl(str(item['id']), 'DOLBY_ATMOS')
                    if self.tool.errmsg != "" or not streamInfo:
                        printErr(14, item['title'] + "(Get Stream Url Err!!" + self.tool.errmsg + ")")
                        continue
                    aAlbumInfo = self.tool.getAlbum(item['album']['id'])
                    fileType = self._getSongExtension(streamInfo['url'])

                    # change targetDir
                    targetDir2 = targetDir
                    if self.config.plfile2arfolder == "True":
                        targetDir2 = self.__creatAlbumDir(aAlbumInfo, self.config.quality)
                        filePath = self.__getAlbumSongSavePath(targetDir2, aAlbumInfo, item, fileType)
                        paraList = {'album': aAlbumInfo, 'title': item['title'], 'trackinfo': item,
                                    'url': streamInfo['url'], 'path': filePath, 'retry': 3, 'key': streamInfo['encryptionKey'], 'codec': streamInfo['codec']}
                    else:
                        seq = self.tool.getIndexStr(index, len(aItemInfo))
                        filePath = targetDir2 + '/' + seq + " "+ pathHelper.replaceLimitChar(item['title'], '-') + fileType
                        paraList = {'album': aAlbumInfo, 'index': index, 'title': item['title'], 'trackinfo': item,
                                    'url': streamInfo['url'], 'path': filePath, 'retry': 3, 'key': streamInfo['encryptionKey'], 'codec': streamInfo['codec']}

                    try:
                        coverPath = targetDir2 + '/' + pathHelper.replaceLimitChar(aAlbumInfo['title'], '-') + '.jpg'
                        coverUrl = self.tool.getAlbumArtworkUrl(aAlbumInfo['cover'])
                        netHelper.downloadFile(coverUrl, coverPath)
                        paraList['coverpath'] = coverPath
                        tmpcoverpath.append(coverPath)
                    except:
                        cmdHelper.myprint("Could not download artwork for '{}'".format(
                            item['title']), cmdHelper.TextColor.Red, None)

                    if self.config.onlym4a == "True":
                        self.check.addPath(filePath.replace(".mp4", ".m4a"))
                    else:
                        self.check.addPath(filePath)
                    self.thread.start(self.__thradfunc_dl, paraList)
                self.thread.waitAll()
                self.tool.removeTmpFile(targetDir)

                # remove cover
                if self.config.savephoto != 'True':
                    for item in tmpcoverpath:
                        pathHelper.remove(item)

                bBreakFlag = True
                bFirstTime = False

                # check
                isErr, errIndex = self.check.checkPaths()
                if isErr:
                    check = printChoice("[Err]\t\t" + str(len(errIndex)) + " Tracks Download Failed.Try Again?(y/n):")
                    if check == 'y' or check == 'Y':
                        bBreakFlag = False

            # download video
            for item in aItemInfo:
                type = item['type']
                item = item['item']
                if type != 'video':
                    continue

                filePath = targetDir + '/' + pathHelper.replaceLimitChar(item['title'], '-') + ".mp4"
                filePath = os.path.abspath(filePath)
                if os.access(filePath, 0):
                    os.remove(filePath)

                videoID = item['id']
                resolutionList, urlList = self.tool.getVideoResolutionList(videoID)
                if urlList is None:
                    printErr(14, item['title'] + '(' + self.tool.errmsg + ')')
                else:
                    selectIndex = self.__getVideoResolutionIndex(resolutionList)
                    if self.ffmpeg.mergerByM3u8_Multithreading2(urlList[int(selectIndex)], filePath, showprogress=self.showpro):
                        printSUCCESS(14, item['title'])
                    else:
                        printErr(14, item['title'] + "(Download Or Merger Err!)")
            if playlist_id is not None:
                return
        return

    def downloadFavorite(self):
        targetDir = self.config.outputdir + "/Favorite/"
        pathHelper.mkdirs(targetDir)

        trackList, videoList = self.tool.getFavorite(self.config.userid)
        if self.tool.errmsg != "":
            printErr(0, "Get FavoriteList Err! " + self.tool.errmsg)
            return

        print("[NumberOfTracks]       %s" % (len(trackList)))
        print("[NumberOfVideos]       %s" % (len(videoList)))
        # download track
        for item in trackList:
            item = item['item']
            streamInfo = self.tool.getStreamUrl(str(item['id']), self.config.quality)
            if self.tool.errmsg != "" or not streamInfo:
                printErr(14, item['title'] + "(Get Stream Url Err!!" + self.tool.errmsg + ")")
                continue

            fileType = self._getSongExtension(streamInfo['url'])
            filePath = targetDir + '/' + pathHelper.replaceLimitChar(item['title'], '-') + fileType
            aAlbumInfo = self.tool.getAlbum(item['album']['id'])
            paraList = {'album': aAlbumInfo, 'title': item['title'], 'trackinfo': item,
                        'url': streamInfo['url'], 'path': filePath, 'retry': 3, 'key': streamInfo['encryptionKey'], 'codec': streamInfo['codec']}
            self.thread.start(self.__thradfunc_dl, paraList)
        self.thread.waitAll()

        # download video
        for item in videoList:
            item = item['item']

            filePath = targetDir + '/' + pathHelper.replaceLimitChar(item['title'], '-') + ".mp4"
            filePath = os.path.abspath(filePath)
            if os.access(filePath, 0):
                os.remove(filePath)

            resolutionList, urlList = self.tool.getVideoResolutionList(item['id'])
            selectIndex = self.__getVideoResolutionIndex(resolutionList)
            if self.ffmpeg.mergerByM3u8_Multithreading2(urlList[int(selectIndex)], filePath, showprogress=self.showpro):
                printSUCCESS(14, item['title'])
            else:
                printErr(14, item['title'])
        return

    def downloadUrl(self, link):
        stype, sid = self.tool.parseLink(link)
        if stype is None or sid is None:
            return
        if stype == "album":
            print("----------------ALBUM------------------")
            self.downloadAlbum(sid)
        elif stype == "track":
            print("----------------TRACK------------------")
            self.downloadTrack(sid)
        elif stype == "video":
            print("----------------VIDEO------------------")
            self.downloadVideo(sid)
        elif stype == "playlist":
            print("--------------PLAYLIST-----------------")
            self.downloadPlaylist(sid)
        elif stype == "artist":
            print("----------------ARTIST-----------------")
            self.downloadArtistAlbum(self.config.includesingle == "True", sid)

    def downloadByFile(self, path):
        if not os.path.exists(path):
            return
        arr = self.tool.parseFile(path)
        print("----------------FILE------------------")
        print("[Number of albums]       %s" % (len(arr['album'])))
        print("[Number of artists]      %s" % (len(arr['artist'])))
        print("[Number of tracks]       %s" % (len(arr['track'])))
        print("[Number of videos]       %s" % (len(arr['video'])))
        print("[Number of URLs]         %s" % (len(arr['url'])))

        if len(arr['album']) > 0:
            redownload = True
            check = printChoice("Skip downloaded files?(y/n):")
            if not cmdHelper.isInputYes(check):
                redownload = False

        for index, item in enumerate(arr['album']):
            print("----Album[{0}/{1}]----".format(index+1, len(arr['album'])))
            print("[ID]          %s" % (item))
            self.downloadAlbum(item, redownload)
        for index, item in enumerate(arr['artist']):
            print(index)
            print("----Artist[{0}/{1}]----".format(index+1, len(arr['artist'])))
            print("[ID]          %s" % (item))
            includeSingles = self.config.includesingle == "True"
            self.downloadArtistAlbum(includeSingles, item)
        for index, item in enumerate(arr['track']):
            print("----Track[{0}/{1}]----".format(index+1, len(arr['track'])))
            print("[ID]                %s" % (item))
            self.downloadTrack(item)
        for index, item in enumerate(arr['video']):
            print("----Video[{0}/{1}]----".format(index+1, len(arr['video'])))
            print("[ID]                %s" % (item))
            self.downloadVideo(item)
        for index, item in enumerate(arr['url']):
            print("----Url[{0}/{1}]----".format(index+1, len(arr['url'])))
            print("[link]        %s" % (item))
            stype, sid = self.tool.parseLink(item)
            if stype is None or sid is None:
                printErr(14, 'Link can`t parse!')
                continue
            print("[ID]          %s" % (sid))
            if stype == "album":
                print("[Type]        %s" % ("album"))
                self.downloadAlbum(sid)
            if stype == "track":
                print("[Type]        %s" % ("track"))
                self.downloadTrack(sid)
            if stype == "video":
                print("[Type]        %s" % ("video"))
                self.downloadVideo(sid)



#!/usr/bin/env python
# coding=utf-8
# -*- encoding: utf-8 -*-
'''
@File    :   tidal.py
@Time    :   2019/02/27
@Author  :   Yaron Huang
@Version :   1.0
@Contact :   yaronhuang@qq.com
@Desc    :   Tidal API
'''
import os
import re
import uuid
import requests
import json
import base64

from aigpy import fileHelper
from aigpy import pathHelper
from aigpy import configHelper
from aigpy import netHelper
from aigpy import systemHelper
from aigpy import tagHelper
from aigpy import configHelper
from aigpy.ffmpegHelper import FFmpegTool
# from tidal_dl import tagHelper
from pydub import AudioSegment
from tidal_dl.printhelper import printWarning

VERSION = '1.9.1'
# VERSION = '2.25.1'
URL_PRE = 'https://api.tidalhifi.com/v1/'
QUALITY = ['HI_RES', 'LOSSLESS', 'HIGH', 'LOW']
TYPE_ARR = ['album', 'track', 'video', 'playlist', 'artist']
RESOLUTION = ['1080', '720', '480', '360', '240']


class TidalTool(object):
    def __init__(self):
        self.config = TidalConfig()
        self.errmsg = ""
        self.tmpfileFlag = 'TIDAL_TMP_'
        self.ffmpeg = FFmpegTool(1)

        self.header = {'X-Tidal-SessionId': self.config.sessionid}
        self.header2 = {'X-Tidal-SessionId': self.config.sessionid2}

        self.header3 = None
        if self.config.accesstoken != '':
            self.header3 = {'authorization': 'Bearer {}'.format(self.config.accesstoken)}

    def _get(self, url, params={}):
        retry = 3

        if self.header3 is not None:
            selectheader = self.header3
        else:
            selectheader = self.header
            if 'soundQuality' in params: 
                if params['soundQuality'] == 'LOSSLESS':
                    selectheader = self.header2

        while retry > 0:
            retry -= 1
            try:
                self.errmsg = ""
                params['countryCode'] = self.config.countrycode
                resp = requests.get(
                    URL_PRE + url,
                    headers=selectheader,
                    params=params).json()
                if 'status' in resp and resp['status'] == 404 and resp['subStatus'] == 2001:
                    self.errmsg = '{}. This might be region-locked.'.format(resp['userMessage'])
                elif 'status' in resp and resp['status'] == 401 and resp['subStatus'] == 4005: #'Asset is not ready for playback'
                    if selectheader != self.header2:
                        selectheader = self.header2
                        continue
                elif 'status' in resp and not resp['status'] == 200:
                    self.errmsg = '{}. Get operation err!'.format(resp['userMessage'])
                    # self.errmsg = "Get operation err!"
                return resp
            except Exception as e:
                if retry <= 0:
                    self.errmsg = 'Function `Http-Get` Err! ' + str(e)
                    return None

    def setTag(self, tag, srcfile, coverpath=None):
        path = pathHelper.getDirName(srcfile)
        name = pathHelper.getFileNameWithoutExtension(srcfile)
        ext = pathHelper.getFileExtension(srcfile)
        oext = ext

        if 'm4a' in ext or 'mp4' in ext:
            oext = '.mp3'
        if 'mp3' not in oext:
            coverpath = None
        tmpfile = path + '/' + 'TMP' + name + oext

        try:
            data = AudioSegment.from_file(srcfile, format=ext[1:])
            check = data.export(tmpfile, format=oext[1:], tags=tag, cover=coverpath)
            check.close()
        except Exception as e:
            pathHelper.remove(tmpfile)
            return

        if fileHelper.getFileSize(tmpfile) > 0:
            pathHelper.remove(srcfile)
            os.rename(tmpfile, path + '/' + name + oext)
        else:
            pathHelper.remove(tmpfile)

    def setTrackMetadata_old(self, track_info, file_path, album_info, index, coverpath):
        tag = {'Artist': track_info['artist']['name'],
               'Album': track_info['album']['title'],
               'Title': track_info['title'],
               'CopyRight': track_info['copyright'],
               'Track': track_info['trackNumber']}
        if index is not None:
            tag['Track'] = str(index)
        if album_info is not None:
            tag['Date'] = album_info['releaseDate']
            tag['Year'] = album_info['releaseDate'].split('-')[0]
        self.setTag(tag, file_path, coverpath)
        return

    def isNeedCovertToM4a(self, file_path):
        if self.config.onlym4a != "True":
            return False
        if '.mp4' not in file_path:
            return False
        if not self.ffmpeg.enable:
            return False
        return True

    def covertMp4toM4a(self, file_path):
        if self.config.onlym4a != "True":
            return file_path
        if '.mp4' not in file_path:
            return file_path
        if not self.ffmpeg.enable:
            return file_path
        new_path = file_path.replace('.mp4', '.m4a')

        pathHelper.remove(new_path)
        if self.ffmpeg.covertFile(file_path, new_path):
            pathHelper.remove(file_path)
            return new_path
        else:
            return file_path

    def _parseContributors(self, roleType, Contributors):
        if Contributors is None:
            return None
        try:
            ret = []
            for item in Contributors['items']:
                if item['role'] == roleType:
                    ret.append(item['name'])
            return ret
        except:
            return None

    def setTrackMetadata(self, track_info, file_path, album_info, index, coverpath, Contributors):
        # isrc,replayGain,releasedate
        obj = tagHelper.TagTool(file_path)
        obj.album = track_info['album']['title']
        obj.title = track_info['title']
        obj.artist = self._getArtists(track_info['artists'])
        obj.copyright = track_info['copyright']
        obj.tracknumber = track_info['trackNumber']
        obj.discnumber = track_info['volumeNumber']
        obj.isrc = track_info['isrc']
        obj.composer = self._parseContributors('Composer', Contributors)
        if index is not None:
            obj.tracknumber = str(index)
        if album_info is not None:
            obj.albumartist = self._getArtists(album_info['artists'])
            obj.date = album_info['releaseDate']
            obj.totaldisc = album_info['numberOfVolumes']
            if obj.totaldisc <= 1:
                obj.totaltrack = album_info['numberOfTracks']
        obj.save(coverpath)
        return

    def removeTmpFile(self, path):
        for root, dirs, files in os.walk(path):
            for name in files:
                if self.tmpfileFlag in name:
                    pathHelper.remove(os.path.join(root, name))

    def getStreamUrl2(self, track_id, quality):
        url = self._get('tracks/' + str(track_id) + '/streamUrl', {'soundQuality': quality})
        if not url:
            resp = self._get('tracks/{}/playbackinfopostpaywall'.format(track_id), {
                 'audioquality': quality,
                 'playbackmode': 'STREAM',
                 'assetpresentation': 'FULL'})
            if resp and 'trackId' in resp:
                #  printWarning(14, "Redirecting: {} -> {}".format(track_id, resp['trackId']))
                track_id = resp['trackId']
                url = self._get('tracks/' + str(track_id) + '/streamUrl', {'soundQuality': quality})
        return url

    def getStreamUrl(self, track_id, quality):
        resp = self._get('tracks/{}/playbackinfopostpaywall'.format(track_id), {
            'audioquality': quality,
            'playbackmode': 'STREAM',
            'assetpresentation': 'FULL'})
        if resp and 'trackId' in resp:
            manifest = json.loads(base64.b64decode(resp['manifest']))
            url = {
                'soundQuality': resp['audioQuality'],
                'codec': manifest['codecs'],
                'url': manifest['urls'][0],
                'encryptionKey': manifest['keyId'] if 'encryptionType' in manifest and manifest['encryptionType'] != 'NONE' else ''
            }
            #  printWarning(14, "Redirecting: {} -> {}".format(track_id, resp['trackId']))
            # track_id = resp['trackId']
            # url = self._get('tracks/' + str(track_id) + '/streamUrl', {'soundQuality': quality})
        return url

    def _getArtists(self, pHash):
        ret = []
        for item in pHash:
            ret.append(item['name'])
        return ret

    def getIndexStr(self, index, sum):
        pre = "0"
        if sum > 99:
            pre = "00"
        if index < 10:
            return pre+str(index) + " "
        if index < 99 and sum > 99:
            return "0"+str(index) + " "
        return str(index) + " "

    def _fixSameTrackName(self, tracks, isOnLayer2=False):
        same = {}
        for item in tracks:
            if isOnLayer2:
                item = item['item']
            if 'version' in item:
                if item['version'] is not None:
                    item['title'] = item['title'] + ' - ' + item['version']
            if item['title'] in same:
                same[item['title']] += 1
            else:
                same[item['title']] = 1
        for item in same:
            if same[item] <= 1:
                continue
            index = 1
            for track in tracks:
                if track:
                    track = track['item']
                if track['title'] != item:
                    continue
                track['title'] += str(index)
                index += 1
        return tracks

    def getPlaylist(self, playlist_id):
        info = self._get('playlists/' + playlist_id)
        if self.errmsg != "":
            return None, None
        list = self.__getItemsList('playlists/' + playlist_id + '/items')
        list = self._fixSameTrackName(list, True)
        return info, list

    def getAlbumTracks(self, album_id):
        info = self._get('albums/' + str(album_id) + '/tracks')
        if self.errmsg != "":
            return info
        # sum = info['totalNumberOfItems']
        for item in info['items']:
            if 'version' in item and item['version'] is not None:
                item['title'] += ' - ' + item['version']
        #     indexs = self._getIndexStr(item['trackNumber'],sum)
        #     item['title'] = indexs + " " + item['title']
        # info['items'] = self._fixSameTrackName(info['items'])
        return info

    def getAlbumVideos(self, album_id):
        # info = self._get('albums/' + str(album_id) + '/items')
        info = self.__getItemsList('albums/' + str(album_id) + '/items')
        if self.errmsg != "":
            return []
        ret = []
        for item in info:
            if item['type'] == 'video':
                ret.append(item)
        return ret

    def getTrack(self, track_id):
        item = self._get('tracks/' + str(track_id))
        if item is None:
            return None
        if 'version' in item and item['version'] is not None:
            item['title'] += ' - ' + item['version']
        return item

    def getAlbum(self, album_id):
        return self._get('albums/' + str(album_id))

    def getVideo(self, video_id):
        return self._get('videos/' + str(video_id))

    def getFavorite(self, user_id):
        trackList = self.__getItemsList('users/' + str(user_id) + '/favorites/tracks')
        tracklist = self._fixSameTrackName(trackList, True)
        videoList = self.__getItemsList('users/' + str(user_id) + '/favorites/videos')
        return trackList, videoList

    def getArtistAlbum(self, artist_id, includeSingles): 
        if includeSingles: 
            items1 = self.__getItemsList('artists/' + str(artist_id) + '/albums',{'filter': 'EPSANDSINGLES'}) 
        else: 
            items1 = []
        # items2 = self.__getItemsList('artists/' + str(artist_id) + '/albums',{'filter': 'COMPILATIONS'})
        items3 = self.__getItemsList('artists/' + str(artist_id) + '/albums')
        itemall = items1 + items3
        return itemall

    def __getItemsList(self, url, in_dirs={}):
        in_dirs['limit'] = 0
        ret = self._get(url, in_dirs)
        count = ret['totalNumberOfItems']
        offset = 0
        limit = 100
        retList = []
        while offset < count:
            in_dirs['limit'] = limit
            in_dirs['offset'] = offset
            items = self._get(url, in_dirs)
            if self.errmsg != "":
                if self.errmsg.find('Too big page') >= 0:
                    limit = limit - 10
                    continue
                else:
                    return retList
            offset = offset + limit
            if retList == None:
                retList = items['items']
            else:
                retList.extend(items['items'])
        return retList

    def getTrackContributors(self, track_id):
        return self._get('tracks/' + str(track_id) + '/contributors')

    def getAlbumArtworkUrl(self, coverid, size=1280):
        if coverid is not None: 
            return 'http://resources.tidal.com/images/{0}/{1}x{1}.jpg'.format(coverid.replace('-', '/'), size) 
        else: 
            return '' 

    def getPlaylistArtworkUrl(self, playlist_uuid, size=1280):
        return 'http://images.tidalhifi.com/im/im?w={1}&h={2}&uuid={0}&rows=2&cols=3&noph'.format(playlist_uuid, size, size)

    def getVideoResolutionList(self, video_id):
        info = self._get('videos/' + str(video_id) + '/streamurl')
        if self.errmsg != "":
            return None, None
        content = netHelper.downloadString(info['url'], None)
        resolutionList, urlList = self.__parseVideoMasterAll(str(content))
        return resolutionList, urlList

    def getVideoMediaPlaylist(self, url):
        urlList = self.__parseVideoMediaPlaylist(url)
        return urlList

    def searchTrack(self, query):
        ret = self._get('search/tracks', {'query': query, 'offset': 0, 'limit': 99})
        return ret

    def __parseVideoMasterAll(self, content):
        pattern = re.compile(r"(?<=RESOLUTION=).+?(?=\\n)")
        resolutionList = pattern.findall(content)
        pattern = re.compile(r"(?<=http).+?(?=\\n)")
        pList = pattern.findall(content)
        urlList = []
        for item in pList:
            urlList.append("http"+item)

        return resolutionList, urlList

    def __parseVideoMediaPlaylist(self, url):
        content = netHelper.downloadString(url, None)
        pattern = re.compile(r"(?<=http).+?(?=\\n)")
        plist = pattern.findall(str(content))
        urllist = []
        for item in plist:
            urllist.append("http"+item)
        return urllist

    def convertAlbumInfoToString(self, aAlbumInfo, aAlbumTracks):
        str = ""
        str += "[ID]          %d\n" % (aAlbumInfo['id'])
        str += "[Title]       %s\n" % (aAlbumInfo['title'])
        str += "[Artists]     %s\n" % (aAlbumInfo['artist']['name'])
        str += "[ReleaseDate] %s\n" % (aAlbumInfo['releaseDate'])
        str += "[SongNum]     %s\n" % (aAlbumInfo['numberOfTracks'])
        str += "[Duration]    %s\n" % (aAlbumInfo['duration'])
        str += '\n'

        i = 0
        while True:
            if i >= int(aAlbumInfo['numberOfVolumes']):
                break
            i = i + 1
            str += "===========Volume %d=============\n" % i
            for item in aAlbumTracks['items']:
                if item['volumeNumber'] != i:
                    continue
                str += '{:<8}'.format("[%d]" % item['trackNumber'])
                str += "%s\n" % item['title']
        return str

    def convertPlaylistInfoToString(seld, aPlaylistInfo, aTrackItems):
        str = ""
        str += "[Title]           %s\n" % (aPlaylistInfo['title'])
        str += "[Type]            %s\n" % (aPlaylistInfo['type'])
        str += "[NumberOfTracks]  %s\n" % (aPlaylistInfo['numberOfTracks'])
        str += "[NumberOfVideos]  %s\n" % (aPlaylistInfo['numberOfVideos'])
        str += "[Duration]        %s\n" % (aPlaylistInfo['duration'])

        i = 0
        str += "===========Track=============\n"
        for item in aTrackItems:
            type = item['type']
            item = item['item']
            if type != 'track':
                continue

            i = i + 1
            str += '{:<8}'.format("[%d]" % i) + item['title'] + '\n'

        i = 0
        str += "\n===========Video=============\n"
        for item in aTrackItems:
            type = item['type']
            item = item['item']
            if type != 'video':
                continue

            i = i + 1
            str += '{:<8}'.format("[%d]" % i) + item['title'] + '\n'
        return str

    def parseLink(self,  link):
        link = link.strip()
        if link.find('http') < 0:
            return None, None

        urlpres = ['tidal.com/', 'tidal.com/browse/']
        for pre in urlpres:
            stype = re.findall(pre + "(.+?)/", link)
            if len(stype) <= 0 or stype[0] not in TYPE_ARR:
                continue
            sid = re.findall(pre+stype[0]+"/(.+)/", link)
            if len(sid) <= 0:
                sid = re.findall(pre+stype[0]+"/(.+)", link)
            if len(sid) <= 0:
                return None, None
            return stype[0], sid[0]
        return None, None

    def parseFile(self, path):
        cfp = configHelper.ParseNoEqual(path)
        ret = cfp
        if 'album' not in ret:
            ret['album'] = []
        if 'artist' not in ret: 
            ret['artist'] = [] 
        if 'track' not in ret:
            ret['track'] = []
        if 'video' not in ret:
            ret['video'] = []
        if 'url' not in ret:
            ret['url'] = []
        return ret

# LogIn and Get SessionID

class TidalToken(object):
    def __init__(self):
        self.clientID = 'ck3zaWMi8Ka_XdI0'
        self.token1 = "MbjR4DLXz1ghC4rV"    
        # self.token2 = "hZ9wuySZCmpLLiui"    # only lossless
        self.token2 = "pl4Vc0hemlAXD0mN"    # only lossless
        try:
            # msg = requests.get( "https://raw.githubusercontent.com/yaronzz/Tidal-Media-Downloader/master/Else/tokens.json", timeout=(20.05, 27.05))
            msg = requests.get( "https://cdn.jsdelivr.net/gh/yaronzz/Tidal-Media-Downloader@latest/Else/tokens.json", timeout=(20.05, 27.05))
            tokens = json.loads(msg.text)
            self.token1 = tokens['token']
            self.token2 = tokens['token_phone']
            if 'clientid' in tokens:
                self.clientid = tokens['clientid']
        except Exception as e:
            pass

class TidalAccount(object):
    def __init__(self, username, password, tokens, bymobile=False, cf=None):
        token = tokens.token1
        if bymobile == True:
            token = tokens.token2
        self.errmsg = ""

        if cf is not None and cf.username == username:
            self.user_id = cf.userid
            self.session_id = cf.sessionid
            if bymobile:
                self.session_id = cf.sessionid2
            if self.validSessionID():
                self.country_code = cf.countrycode
                self.password = password
                return

        self.username = username
        self.token = token
        self.uuid = str(uuid.uuid4()).replace('-', '')[16:]
        self.getSessionID(password)

    def getSessionID(self, password):
        postParams = {
            'username': self.username,
            'password': password,
            'token': self.token,
            'clientUniqueKey': self.uuid,
            'clientVersion': VERSION,
        }
        re = requests.post(URL_PRE + 'login/username', data=postParams).json()
        if 'status' in re:
            label = ''
            if 'userMessage' in re and re['userMessage'] is not None:
                label = re['userMessage']
            if re['status'] == 401:
                self.errmsg = "Username or password err!" + label
            else:
                self.errmsg = "Get sessionid err!" + label
        else:
            self.session_id = re['sessionId']
            self.user_id = re['userId']
            self.country_code = re['countryCode']
            if self.validSessionID() is False:
                self.errmsg = "Sessionid is unvalid!"
    
    def validSessionID(self):
        ret = requests.get(URL_PRE + 'users/' + str(self.user_id), params={'sessionId': self.session_id}).json()
        if 'status' in ret and not ret['status'] == 200:
                return False
        return True

# Config Tool
class TidalConfig(object):
    FILE_NAME = "tidal-dl.ini"
    AES_KEY = "hhxx2020TTXS"
    def __init__(self):
        self.outputdir = configHelper.GetValue("base", "outputdir", "./", self.FILE_NAME)
        self.sessionid = configHelper.GetValue("base", "sessionid", "", self.FILE_NAME)
        self.countrycode = configHelper.GetValue("base", "countrycode", "", self.FILE_NAME)
        self.quality = configHelper.GetValue("base", "quality", "LOSSLESS", self.FILE_NAME)
        self.resolution = configHelper.GetValue("base", "resolution", "720", self.FILE_NAME)
        self.username = configHelper.GetValue("base", "username", "", self.FILE_NAME)
        self.password = configHelper.GetValue("base", "password", "", self.FILE_NAME, aesKey=self.AES_KEY)
        self.userid = configHelper.GetValue("base", "userid", "", self.FILE_NAME)
        self.threadnum = configHelper.GetValue("base", "threadnum", "1", self.FILE_NAME)
        self.sessionid2 = configHelper.GetValue("base", "sessionid2", "", self.FILE_NAME)
        self.onlym4a = configHelper.GetValue("base", "onlym4a", "False", self.FILE_NAME)
        self.showprogress = configHelper.GetValue("base", "showprogress", "True", self.FILE_NAME)
        self.addhyphen = configHelper.GetValue("base", "addhyphen", "False", self.FILE_NAME)
        self.addyear = configHelper.GetValue("base", "addyear", "No", self.FILE_NAME)
        self.plfile2arfolder = configHelper.GetValue("base", "plfile2arfolder", "False", self.FILE_NAME)
        self.addexplicit = configHelper.GetValue("base", "addexplicit", "False", self.FILE_NAME)
        self.includesingle = configHelper.GetValue("base", "includesingle", "True", self.FILE_NAME)
        self.savephoto = configHelper.GetValue("base", "savephoto", "True", self.FILE_NAME)
        self.lastlogintime = configHelper.GetValue("base", "lastlogintime", "", self.FILE_NAME)
        self.accesstoken = configHelper.GetValue("base", "accesstoken", "", self.FILE_NAME)

    def set_lastlogintime(self, status):
        self.lastlogintime = status
        configHelper.SetValue("base", "lastlogintime", status, self.FILE_NAME)

    def set_savephoto(self, status):
        if status == 0:
            self.savephoto = "False"
        else:
            self.savephoto = "True"
        configHelper.SetValue("base", "savephoto", self.savephoto, self.FILE_NAME)

    def set_includesingle(self, status):
        if status == 0:
            self.includesingle = "False"
        else:
            self.includesingle = "True"
        configHelper.SetValue("base", "includesingle", self.includesingle, self.FILE_NAME)

    def set_plfile2arfolder(self, status):
        if status == 0:
            self.plfile2arfolder = "False"
        else:
            self.plfile2arfolder = "True"
        configHelper.SetValue("base", "plfile2arfolder", self.plfile2arfolder, self.FILE_NAME)

    def set_onlym4a(self, status):
        if status == 0:
            self.onlym4a = "False"
        else:
            self.onlym4a = "True"
        configHelper.SetValue("base", "onlym4a", self.onlym4a, self.FILE_NAME)

    def set_showprogress(self, status):
        if status == 0:
            self.showprogress = "False"
        else:
            self.showprogress = "True"
        configHelper.SetValue("base", "showprogress", self.showprogress, self.FILE_NAME)

    def set_addhyphen(self, status):
        if status == 0:
            self.addhyphen = "False"
        else:
            self.addhyphen = "True"
        configHelper.SetValue("base", "addhyphen", self.addhyphen, self.FILE_NAME)

    def set_addyear(self, status):
        self.addyear = status
        configHelper.SetValue("base", "addyear", status, self.FILE_NAME)

    def set_addexplicit(self, status):
        if status == 0:
            self.addexplicit = "False"
        else:
            self.addexplicit = "True"
        configHelper.SetValue("base", "addexplicit", self.addexplicit, self.FILE_NAME)

    def set_threadnum(self, threadnum):
        self.threadnum = threadnum
        configHelper.SetValue("base", "threadnum", threadnum, self.FILE_NAME)

    def set_outputdir(self, outputdir):
        self.outputdir = outputdir
        configHelper.SetValue("base", "outputdir", outputdir, self.FILE_NAME)

    def set_quality(self, quality):
        self.quality = quality
        configHelper.SetValue("base", "quality", quality, self.FILE_NAME)

    def set_resolution(self, resolution):
        self.resolution = resolution
        configHelper.SetValue("base", "resolution", resolution, self.FILE_NAME)

    def set_account(self, username, password, sessionid, countrycode, userid, sessionid2):
        self.username = username
        self.password = password
        self.sessionid = sessionid
        self.sessionid2 = sessionid2
        self.countrycode = countrycode
        self.userid = userid
        configHelper.SetValue("base", "username", username, self.FILE_NAME)
        configHelper.SetValue("base", "password", password, self.FILE_NAME, aesKey=self.AES_KEY)
        configHelper.SetValue("base", "sessionid", sessionid, self.FILE_NAME)
        configHelper.SetValue("base", "sessionid2", sessionid2, self.FILE_NAME)
        configHelper.SetValue("base", "countrycode", countrycode, self.FILE_NAME)
        configHelper.SetValue("base", "userid", str(userid), self.FILE_NAME)

    def set_account2(self, username, password, accesstoken, countrycode, userid):
        self.username = username
        self.password = password
        self.accesstoken = accesstoken
        self.countrycode = countrycode
        self.userid = userid
        configHelper.SetValue("base", "username", username, self.FILE_NAME)
        configHelper.SetValue("base", "password", password, self.FILE_NAME, aesKey=self.AES_KEY)
        configHelper.SetValue("base", "accesstoken", accesstoken, self.FILE_NAME)
        configHelper.SetValue("base", "countrycode", countrycode, self.FILE_NAME)
        configHelper.SetValue("base", "userid", str(userid), self.FILE_NAME)

    def valid_quality(self, quality):
        if quality in QUALITY:
            return True
        return False

    def valid_threadnum(self, threadnum):
        try:
            num = int(threadnum)
            return num > 0
        except:
            return False


# if __name__ == '__main__':
#     tool = TidalTool()
#     tool.getArtistAlbum('3529466')



# -*- coding: utf-8 -*-
import sys
import os
import time

from aigpy import pipHelper
from aigpy import pathHelper
from aigpy.cmdHelper import myinput, myinputInt

from tidal_dl.session import TidalMobileSession
from tidal_dl.tidal import TidalConfig
from tidal_dl.tidal import TidalAccount, TidalToken
from tidal_dl.download import Download
from tidal_dl.printhelper import printMenu, printChoice2, printErr, printWarning, LOG

TIDAL_DL_VERSION = "2020.6.14.0"
TIDAL_TOKEN = TidalToken()

def logInByTime():
    cf = TidalConfig()
    try:
        Lasttime = float(cf.lastlogintime)
        Lastlocaltime = time.localtime(Lasttime)
    except:
        Lastlocaltime = None
    
    Curtime = time.time()
    Curlocaltime = time.localtime(Curtime)
    if Lastlocaltime is not None:
        if Curlocaltime.tm_mon == Lastlocaltime.tm_mon and Curlocaltime.tm_mday == Lastlocaltime.tm_mday and Curlocaltime.tm_hour < Lastlocaltime.tm_hour + 3:
            return
        
    if logIn(cf.username, cf.password) == False:
        while logIn("", "") == False:
            pass

def logIn(username="", password=""):
    cf = TidalConfig()
    if username == "" or password == "":
        print("----------------LogIn------------------")
        username = myinput("username:")
        password = myinput("password:")

    account3 = TidalMobileSession(username, password, TIDAL_TOKEN.clientID, cf)
    if account3.errmsg is None:
        cf.set_account2(username, password, account3.access_token, account3.country_code, account3.user_id)
    else:
        cf.set_account2(username, password, '', '', '')
        printWarning(0, 'Login err(by mobile)!' + account3.errmsg)

        account = TidalAccount(username, password, TIDAL_TOKEN, False, cf)
        account2 = TidalAccount(username, password, TIDAL_TOKEN, True, cf)
        if account.errmsg != "" and account2.errmsg != "":
            printErr(0, account.errmsg)
            return False
        elif account.errmsg != "":
            account = account2
        elif account2.errmsg != "":
            account2 = account
        cf.set_account(username, password, account.session_id, account.country_code, account.user_id, account2.session_id)

    
    Curtime = time.time()
    cf.set_lastlogintime(str(Curtime))
    return True


def showConfig():
    cf = TidalConfig()
    print("----------------Config------------------")
    print("Username                         : " + cf.username)
    print("Output directory                 : " + cf.outputdir)
    print("SessionID                        : " + cf.sessionid)
    print("Country Code                     : " + cf.countrycode)
    print("Sound Quality                    : " + cf.quality)
    print("Video Resolution                 : " + cf.resolution)
    print("Download Threads                 : " + cf.threadnum)
    print("Only M4a                         : " + cf.onlym4a)
    print("Show download progress           : " + cf.showprogress + "(enable when threadnum=1)")
    print("Use hyphens                      : " + cf.addhyphen + "(between number and title)")
    print("Add year                         : " + cf.addyear + "(in album title)")
    print("Add explicit tag                 : " + cf.addexplicit)
    print("Playlist songs in artist folders : " + cf.plfile2arfolder + "(organized with artist folder)")
    print("Include singles                  : " + cf.includesingle + "(download artist album)")
    print("Save covers                      : " + cf.savephoto)
    print("Version                          : " + TIDAL_DL_VERSION)
    myinput("Enter to return.")


def setting():
    cf = TidalConfig()
    print("----------------Settings----------------")
    print("Output directory                 :\t" + cf.outputdir)
    print("Sound Quality                    :\t" + cf.quality)
    print("Video Resolution                 :\t" + cf.resolution)
    print("Download Threads                 :\t" + cf.threadnum)
    print("Only M4a                         :\t" + cf.onlym4a)
    print("Show download progress           :\t" + cf.showprogress + "(enable when threadnum=1)")
    print("Use hyphens                      :\t" + cf.addhyphen + "(between number and title)")
    print("Add year                         :\t" + cf.addyear + "(in album title)")
    print("Add explicit tag                 :\t" + cf.addexplicit)
    print("Playlist songs in artist folders :\t" + cf.plfile2arfolder + "(organized with artist folder)")
    print("Include singles                  :\t" + cf.includesingle + "(download artist album)")
    print("Save covers                      : " + cf.savephoto)
    while True:
        outputdir = myinput("Output directory(Enter '0' Unchanged):".ljust(12))
        if outputdir == '0':
            outputdir = cf.outputdir
            break
        if os.path.isdir(outputdir) == False:
            printErr(0, "Path is Err!")
            continue
        break
    while True:
        index = myinputInt("Download Quality(0-LOW,1-HIGH,2-LOSSLESS,3-HI_RES):".ljust(12), 999)
        if index > 3 or index < 0:
            printErr(0, "Quality Err!")
            continue
        if index == 0:
            quality = 'LOW'
        if index == 1:
            quality = 'HIGH'
        if index == 2:
            quality = 'LOSSLESS'
        if index == 3:
            quality = 'HI_RES'
        break
    while True:
        index = myinputInt("Video resolution(0-1080,1-720,2-480,3-360,4-240):".ljust(12), 99)
        if index > 4 or index < 0:
            printErr(0, "Resolution Err")
            continue
        if index == 0:
            resolution = '1080'
        if index == 1:
            resolution = '720'
        if index == 2:
            resolution = '480'
        if index == 3:
            resolution = '360'
        if index == 4:
            resolution = '240'
        break
    while True:
        threadnum = myinput("Number of download threads:".ljust(12))
        if cf.valid_threadnum(threadnum) == False:
            printErr(0, "ThreadNum Err")
            continue
        break

    status = myinputInt("Convert Mp4 to M4a(0-No, 1-Yes):".ljust(12), 0)
    status2 = myinputInt("Show download progress (only available on single thread)(0-No, 1-Yes):".ljust(12), 0)
    status3 = myinputInt("Use hyphens instead of spaces in file names(0-No, 1-Yes):".ljust(12), 0)

    while True:
        index = myinputInt("Add year to album folder names(0-No, 1-Before, 2-After):".ljust(12), 99)
        if index > 2 or index < 0:
            printErr(0, "Addyear input Err")
            continue
        if index == 0:
            addyear = 'No'
        if index == 1:
            addyear = 'Before'
        if index == 2:
            addyear = 'After'
        break

    status5 = myinputInt("Download playlist songs in artist folder structure? (0-No,1-Yes):".ljust(12), 0)
    status6 = myinputInt("Add explicit tag to file names(0-No, 1-Yes):".ljust(12), 0)
    status7 = myinputInt("Include singles and EPs when downloading an artist's albums (0-No, 1-Yes):".ljust(12), 0)
    status8 = myinputInt("Save covers(0-No, 1-Yes):".ljust(12), 0)

    cf.set_outputdir(outputdir)
    cf.set_quality(quality)
    cf.set_resolution(resolution)
    cf.set_threadnum(threadnum)
    cf.set_onlym4a(status)
    cf.set_showprogress(status2)
    cf.set_addhyphen(status3)
    cf.set_addyear(addyear)
    cf.set_plfile2arfolder(status5)
    cf.set_addexplicit(status6)
    cf.set_includesingle(status7)
    cf.set_savephoto(status8)

    pathHelper.mkdirs(outputdir + "/Album/")
    pathHelper.mkdirs(outputdir + "/Playlist/")
    pathHelper.mkdirs(outputdir + "/Video/")
    pathHelper.mkdirs(outputdir + "/Favorite/")
    return


def main(argv=None):
    if byCommand() is True:
        return

    print(LOG)
    cf = TidalConfig()
    if logIn(cf.username, cf.password) == False:
        while logIn("", "") == False:
            pass

    cf = TidalConfig()
    onlineVer = pipHelper.getLastVersion('tidal-dl')
    print("====================Tidal-dl========================")
    print("Output directory                 : " + cf.outputdir)
    print("Sound Quality                    : " + cf.quality)
    print("Video Resolution                 : " + cf.resolution)
    print("Download Threads                 : " + cf.threadnum)
    print("Only M4a                         : " + cf.onlym4a)
    print("Show download progress           : " + cf.showprogress + "(enable when threadnum=1)")
    print("Use hyphens                      : " + cf.addhyphen + "(between number and title)")
    print("Add year                         : " + cf.addyear + "(in album title)")
    print("Add explicit tag                 : " + cf.addexplicit)
    print("Playlist songs in artist folders : " + cf.plfile2arfolder + "(organized with artist folder)")
    print("Include singles                  : " + cf.includesingle + "(download artist album)")
    print("Save covers                      : " + cf.savephoto)
    print("Current Version                  : " + TIDAL_DL_VERSION)
    if onlineVer != None:
        print("Latest Version                   : " + onlineVer)
    print("====================================================")

    dl = Download(cf.threadnum)
    if not dl.ffmpeg.enable:
        printWarning(0, "Couldn't find ffmpeg!\n")
    while True:
        printMenu()
        strchoice, choice = printChoice2("Enter Choice:", 99)
        if choice == 0:
            return
        elif choice == 1:
            logIn()
            cf = TidalConfig()
            dl = Download(cf.threadnum)
        elif choice == 2:
            setting()
            cf = TidalConfig()
            dl = Download(cf.threadnum)
        elif choice == 3:
            dl.downloadAlbum()
        elif choice == 4:
            dl.downloadTrack()
        elif choice == 5:
            dl.downloadPlaylist()
        elif choice == 6:
            dl.downloadVideo()
        elif choice == 7:
            dl.downloadFavorite()
        elif choice == 8:
            dl.downloadArtistAlbum(cf.includesingle == "True")
        elif choice == 9: 
            showConfig()
        #Hidden Code For Developer [200-299]
        elif choice == 200:
            dl.downloadArtistAlbum(False)
        else:
            dl.downloadUrl(strchoice)
            dl.downloadByFile(strchoice)

def byCommand():
    try:
        if len(sys.argv) != 2:
            return False
        logInByTime()
        cf = TidalConfig()
        dl = Download(cf.threadnum)
        dl.downloadUrl(sys.argv[1])
        return True
    except Exception as e:
        return False
    return False

def debug():
    cf = TidalConfig()

    # import tidalapi
    # test = tidalapi.Config()
    # session = tidalapi.Session()
    # session.login(cf.username, cf.password)
    # TIDAL_TOKEN.token2 = "wc8j_yBJd20zOmx0"
    # x-tidal-token: qe5mgUGPtIfbgN574ngS74Sd1OmKIfvcLx7e28Yk
    # TIDAL_TOKEN.token1 = "CzET4vdadNUFQ5JU"
    
    # TIDAL_TOKEN.token1 = "u5qPNNYIbD0S0o36MrAiFZ56K6qMCrCmYPzZuTnV"
    # TIDAL_TOKEN.token1 = TIDAL_TOKEN.token2
    # account = TidalAccount(cf.username, cf.password, TIDAL_TOKEN)

    # import requests
    # import uuid
    # from urllib.parse import urljoin

    # headers = {"X-Tidal-Token": 'u5qPNNYIbD0S0o36MrAiFZ56K6qMCrCmYPzZuTnV'}
    # postParams = {
    #     'username': 'XXX',
    #     'password': 'XXX',
    #     'token': 'u5qPNNYIbD0S0o36MrAiFZ56K6qMCrCmYPzZuTnV',
    #     'clientUniqueKey': str(uuid.uuid4()).replace('-', '')[16:],
    #     'clientVersion': '1.9.1'
    # }
    # location = 'https://api.tidalhifi.com/v1/'
    # myurl = urljoin(location, 'login/username')
    # re = requests.post(myurl, data=postParams)
    tt = TidalMobileSession(cf.username, cf.password,'RnhXoTmoJgARtXHr')
    if logIn(cf.username, cf.password) == False:
        pass

    # add tag Credits,Info song and full tag (discnumber,irsc,composer,arrenger,publisher,replayGain,releasedate)
    # https://api.tidal.com/v1/albums/71121869/tracks?token=wdgaB1CilGA-S_s2&countryCode=TH
    print('\nThis is the debug version!!\n')
    # os.system("pip install aigpy --upgrade")
    # trackid = 70973230
    dl = Download(1)
    # dl.downloadTrack("90521281")
    dl.downloadTrack("131069665") #Dolby Atmos
    # https://tidal.com/browse/track/139322230 360
    # dl.downloadAlbum("120929182", True)
    # dl.tool.getPlaylist("36ea71a8-445e-41a4-82ab-6628c581535d")
    # ss = dl.tool.getPlaylistArtworkUrl("36ea71a8-445e-41a4-82ab-6628c581535d")
    # ss = dl.tool.getPlaylistArtworkUrl("36ea71a8-445e-41a4-82ab-6628c581535d",480)
    # dl.downloadVideo(57261945) #1hours
    # tidal.com/browse/track/125155002 dubi
    # dl.downloadVideo(84094460)
    # https://tidal.com/browse/track/70973230

# if __name__ == '__main__':
#     main(sys.argv)


__all__ = ['debug', 'main', 'tidal', 'download']



#!/usr/bin/env python
# -*- encoding: utf-8 -*-
import pickle
import uuid
import os
import json
import urllib.parse as urlparse
from urllib.parse import parse_qs
import hashlib
import base64
import secrets
from datetime import datetime, timedelta

import requests
from urllib3.util.retry import Retry
from requests.adapters import HTTPAdapter
from tidal_dl.tidal import TidalConfig

'''
@File    :   session.py
@Time    :   2020/06/09
@Author  :   Yaronzz
@Version :   1.0
@Contact :   yaronhuang@foxmail.com
@Desc    :   Class from project "Redsea"
'''

class TidalMobileSession(object):
    # client_id	CLGLE93UA5gm42Og  ck3zaWMi8Ka_XdI0
    # client_id	RnhXoTmoJgARtXHr
    def __init__(self, username, password, client_id='ck3zaWMi8Ka_XdI0', cf=None):
        self.username = username
        self.client_id = client_id
        self.redirect_uri =  'https://tidal.com/android/login/auth'
        self.code_verifier = base64.urlsafe_b64encode(secrets.token_bytes(32)).rstrip(b'=')
        self.code_challenge = base64.urlsafe_b64encode(hashlib.sha256(self.code_verifier).digest()).rstrip(b'=')
        self.client_unique_key = secrets.token_hex(16)
        self.access_token = None
        self.user_id = None
        self.country_code = None
        self.errmsg = None

        if cf is not None and cf.username == username:
            self.user_id = cf.userid
            self.access_token = cf.accesstoken
            if self.valid():
                self.country_code = cf.countrycode
                self.password = password
                return
        self.auth(password)

    def auth(self, password):
        s = requests.Session()

        params = {
            'response_type': 'code',
            'redirect_uri': self.redirect_uri,
            'lang': 'en_US',
            'appMode': 'android',
            'client_id': self.client_id,
            'client_unique_key': self.client_unique_key,
            'code_challenge': self.code_challenge,
            'code_challenge_method': 'S256'
        }

        # retrieve csrf token for subsequent request
        r = s.get('https://login.tidal.com/authorize', params=params)

        # enter email, verify email is valid
        r = s.post('https://login.tidal.com/email', params=params, json={
            '_csrf': s.cookies['token'],
            'email': self.username,
            'recaptchaResponse': ''
        })
        if r.status_code != 200:
            self.errmsg = 'Unknown in https://login.tidal.com/email'
            return
        if not r.json()['isValidEmail']:
            self.errmsg = 'Invalid email'
            return
        if r.json()['newUser']:
            self.errmsg = 'User does not exist'
            return

        # login with user credentials
        r = s.post('https://login.tidal.com/email/user/existing', params=params, json={
            '_csrf': s.cookies['token'],
            'email': self.username,
            'password': password
        })
        if r.status_code != 200:
            self.errmsg = 'Unknown in https://login.tidal.com/email/user/existing'
            return

        # retrieve access code
        r = s.get('https://login.tidal.com/success?lang=en', allow_redirects=False)
        if r.status_code == 401:
            self.errmsg = 'Incorrect password'
            return
        if r.status_code != 302:
            self.errmsg = 'Unknown in https://login.tidal.com/success?lang=en'
            return
        url = urlparse.urlparse(r.headers['location'])
        oauth_code = parse_qs(url.query)['code'][0]

        # exchange access code for oauth token
        r = requests.post('https://auth.tidal.com/v1/oauth2/token', data={
            'code': oauth_code,
            'client_id': self.client_id,
            'grant_type': 'authorization_code',
            'redirect_uri': self.redirect_uri,
            'scope': 'r_usr w_usr w_sub',
            'code_verifier': self.code_verifier,
            'client_unique_key': self.client_unique_key
        })
        if r.status_code != 200:
            self.errmsg = 'Unknown in https://auth.tidal.com/v1/oauth2/token'
            return
        self.access_token = r.json()['access_token']

        r = requests.get('https://api.tidal.com/v1/sessions', headers=self.auth_headers())
        if r.status_code != 200:
            self.errmsg = 'Unknown in https://api.tidal.com/v1/sessions'
            return
        self.user_id = r.json()['userId']
        self.country_code = r.json()['countryCode']

    def valid(self):
        r = requests.get('https://api.tidal.com/v1/sessions', headers=self.auth_headers())
        return r.status_code == 200

    def auth_headers(self):
        return {'authorization': 'Bearer {}'.format(self.access_token)}

